package util;

import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.PrivateKey;
import java.security.SecureRandom;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import javax.xml.bind.DatatypeConverter;

import org.apache.commons.io.IOUtils;
import org.bouncycastle.bcpg.ArmoredOutputStream;
import org.bouncycastle.openpgp.PGPEncryptedData;
import org.bouncycastle.openpgp.PGPEncryptedDataGenerator;
import org.bouncycastle.openpgp.PGPEncryptedDataList;
import org.bouncycastle.openpgp.PGPException;
import org.bouncycastle.openpgp.PGPLiteralData;
import org.bouncycastle.openpgp.PGPLiteralDataGenerator;
import org.bouncycastle.openpgp.PGPObjectFactory;
import org.bouncycastle.openpgp.PGPPrivateKey;
import org.bouncycastle.openpgp.PGPPublicKey;
import org.bouncycastle.openpgp.PGPPublicKeyEncryptedData;
import org.bouncycastle.openpgp.PGPPublicKeyRing;
import org.bouncycastle.openpgp.PGPPublicKeyRingCollection;
import org.bouncycastle.openpgp.PGPSecretKey;
import org.bouncycastle.openpgp.PGPSecretKeyRing;
import org.bouncycastle.openpgp.PGPSecretKeyRingCollection;
import org.bouncycastle.openpgp.PGPUtil;
import org.bouncycastle.openpgp.operator.PublicKeyDataDecryptorFactory;
import org.bouncycastle.openpgp.operator.jcajce.JcaKeyFingerprintCalculator;
import org.bouncycastle.openpgp.operator.jcajce.JcePBESecretKeyDecryptorBuilder;
import org.bouncycastle.openpgp.operator.jcajce.JcePGPDataEncryptorBuilder;
import org.bouncycastle.openpgp.operator.jcajce.JcePublicKeyDataDecryptorFactoryBuilder;
import org.bouncycastle.openpgp.operator.jcajce.JcePublicKeyKeyEncryptionMethodGenerator;
import org.bouncycastle.util.encoders.Base64;
import org.bouncycastle.util.io.Streams;

import io.jsonwebtoken.Jwts;
import io.jsonwebtoken.SignatureAlgorithm;

public class NewPGPUtil {

	public static PGPPublicKey readPublicKey(String pgpPublicKeyAsString) throws IOException, PGPException {
		InputStream keyIn = new ByteArrayInputStream(pgpPublicKeyAsString.getBytes());
		PGPPublicKey pubKey = readPublicKey(new ByteArrayInputStream(pgpPublicKeyAsString.getBytes()));
		keyIn.close();
		return pubKey;
	}

	/**
	 * A simple routine that opens a key ring file and loads the first available key
	 * suitable for encryption.
	 *
	 * @param input data stream containing the public key data
	 * @return the first public key found.
	 */
	public static PGPPublicKey readPublicKey(InputStream input) throws IOException, PGPException {
		PGPPublicKeyRingCollection pgpPub = new PGPPublicKeyRingCollection(PGPUtil.getDecoderStream(input),
				new JcaKeyFingerprintCalculator());

		//
		// we just loop through the collection till we find a key suitable for
		// encryption, in the real
		// world you would probably want to be a bit smarter about this.
		//

		Iterator keyRingIter = pgpPub.getKeyRings();
		while (keyRingIter.hasNext()) {
			PGPPublicKeyRing keyRing = (PGPPublicKeyRing) keyRingIter.next();

			Iterator keyIter = keyRing.getPublicKeys();
			while (keyIter.hasNext()) {
				PGPPublicKey key = (PGPPublicKey) keyIter.next();
				if (key.isEncryptionKey() && !key.isMasterKey()) {
					return key;
				}
			}
		}

		throw new IllegalArgumentException("Can't find encryption key in key ring.");
	}

	public static String encrpytRequestPayload(String requestPayloadPlainString, List<PGPPublicKey> publicKeys,
			boolean armor, boolean withIntegrityCheck, boolean isMasterKey) throws IOException, PGPException {
		ByteArrayOutputStream out = new ByteArrayOutputStream();
		OutputStream opOut;
		if (armor) {
			opOut = new ArmoredOutputStream(out);
		} else {
			opOut = out;
		}

		ByteArrayOutputStream byteArray = new ByteArrayOutputStream();

		InputStream inputStream = IOUtils.toInputStream(requestPayloadPlainString);

		PGPLiteralDataGenerator literalDataGenerator = new PGPLiteralDataGenerator();
		OutputStream outputStream = literalDataGenerator.open(byteArray, PGPLiteralData.BINARY, PGPLiteralData.CONSOLE,
				inputStream.available(), new Date());
		Streams.pipeAll(inputStream, outputStream);
		outputStream.close();

		JcePGPDataEncryptorBuilder dataEncryptorBuilder = new JcePGPDataEncryptorBuilder(PGPEncryptedData.AES_256)
				.setWithIntegrityPacket(withIntegrityCheck).setSecureRandom(new SecureRandom()).setProvider("BC");

		PGPEncryptedDataGenerator encryptedDataGenerator = new PGPEncryptedDataGenerator(dataEncryptorBuilder);
		PGPPublicKey encryptionKey = null;

		Iterator keyIter = publicKeys.iterator();
		while (keyIter.hasNext()) {
			PGPPublicKey key = (PGPPublicKey) keyIter.next();
			if (key.isEncryptionKey() && (!key.isMasterKey() || isMasterKey)) {
				encryptionKey = key;
			}
		}
		if (encryptionKey == null) {
			throw new IllegalStateException("matching encrypted data not found");
		}
		JcePublicKeyKeyEncryptionMethodGenerator encryptionMethodGenerator = new JcePublicKeyKeyEncryptionMethodGenerator(
				encryptionKey).setProvider("BC").setSecureRandom(new SecureRandom());

		encryptedDataGenerator.addMethod(encryptionMethodGenerator);

		byte[] bytes = byteArray.toByteArray();
		try (OutputStream dbOut = encryptedDataGenerator.open(opOut, bytes.length)) {
			dbOut.write(bytes);
			dbOut.close();
			opOut.close();

			return Base64.toBase64String(out.toByteArray());

		}
	}

	// Helper method for getting message digest
	private static String generateMessageDigest(String message, String hashingAlgorithm)
			throws NoSuchAlgorithmException {
		return DatatypeConverter.printHexBinary(MessageDigest.getInstance(hashingAlgorithm).digest(message.getBytes()))
				.toLowerCase();
	}

	/**
	 * createToken
	 *
	 * @param profileId                           profileId
	 * @param requestPayloadEncryptedBase64String requestPayloadEncryptedBase64String
	 * @param keyId                               keyId
	 * @param privateKey                          privateKey
	 * @return token
	 * @throws NoSuchAlgorithmException NoSuchAlgorithmException
	 */
	public static String createToken(String profileId, String requestPayloadEncryptedBase64String, String keyId,
			PrivateKey privateKey) throws NoSuchAlgorithmException {

		String hashingAlgorithm = "RSASHA256";
		// Generate unique ID for this API call
		String jwtId = UUID.randomUUID().toString();
		// Build JWT
		long nowSeconds = System.currentTimeMillis() / 1000;
		Map<String, Object> claims = new HashMap<String, Object>();
		claims.put("sub", profileId);
		claims.put("iat", nowSeconds);
		claims.put("aud", "HSBC.MKT");
		claims.put("jti", jwtId);
		String payloadHash = generateMessageDigest(requestPayloadEncryptedBase64String, "SHA-256");
		claims.put("payload_hash", payloadHash);
		System.out.println("payload_hash:" + payloadHash);
		claims.put("payload_hash_alg", hashingAlgorithm);

		Map<String, Object> header = new HashMap<String, Object>();
		header.put("typ", "JWT");
		header.put("kid", keyId);
		header.put("alg", SignatureAlgorithm.RS256.getValue());
		header.put("ver", "1.0");
		String token = Jwts.builder().setHeader(header).setClaims(claims).signWith(SignatureAlgorithm.RS256, privateKey)
				.compact();
		return token;
	}

	public static PGPPrivateKey readSecretKey(String keyId, String privateKeyAsString, String passphrase)
			throws IOException, PGPException {
		InputStream keyIn = new ByteArrayInputStream(privateKeyAsString.getBytes());
		PGPSecretKey secKey = readSecretKey(keyIn);
		keyIn.close();
		PGPPrivateKey key = secKey.extractPrivateKey(
				new JcePBESecretKeyDecryptorBuilder().setProvider("BC").build(passphrase.toCharArray()));
		return key;
	}

	/**
	 * A simple routine that opens a key ring file and loads the first available key
	 * suitable for signature generation.
	 *
	 * @param input stream to read the secret key ring collection from.
	 * @return a secret key.
	 * @throws IOException  on a problem with using the input stream.
	 * @throws PGPException if there is an issue parsing the input stream.
	 */
	public static PGPSecretKey readSecretKey(InputStream input) throws IOException, PGPException {
		PGPSecretKeyRingCollection pgpSec = new PGPSecretKeyRingCollection(PGPUtil.getDecoderStream(input),
				new JcaKeyFingerprintCalculator());

		//
		// we just loop through the collection till we find a key suitable for
		// encryption, in the real
		// world you would probably want to be a bit smarter about this.
		//

		Iterator keyRingIter = pgpSec.getKeyRings();
		while (keyRingIter.hasNext()) {
			PGPSecretKeyRing keyRing = (PGPSecretKeyRing) keyRingIter.next();

			Iterator keyIter = keyRing.getSecretKeys();
			while (keyIter.hasNext()) {
				PGPSecretKey key = (PGPSecretKey) keyIter.next();

				if (key.isSigningKey() && !key.isMasterKey()) {
					return key;
				}
			}
		}

		throw new IllegalArgumentException("Can't find signing key in key ring.");
	}

	public static PGPPrivateKey findSecretKey(PGPSecretKeyRingCollection pgpSec, long keyID, char[] pass)
			throws PGPException, NoSuchProviderException {
		PGPSecretKey pgpSecKey = pgpSec.getSecretKey(keyID);

		if (pgpSecKey == null) {
			return null;
		}

		return pgpSecKey.extractPrivateKey(new JcePBESecretKeyDecryptorBuilder().setProvider("BC").build(pass));
	}

	public static String decryptResponseMessage(PGPPrivateKey privateKey, String responseMessageEncryptedBase64String)
			throws PGPException, IOException {

		JcaKeyFingerprintCalculator keyFingerPrintCalculator = new JcaKeyFingerprintCalculator();

		String epicStrNew = new String(Base64.decode(responseMessageEncryptedBase64String));

		BufferedInputStream dataStream = new BufferedInputStream(IOUtils.toInputStream(epicStrNew));

		InputStream decoderStream = PGPUtil.getDecoderStream(dataStream);

		PGPObjectFactory pgpObjectFactory = new PGPObjectFactory(decoderStream, keyFingerPrintCalculator);

		PGPEncryptedDataList pgpEncryptedDataList = (PGPEncryptedDataList) pgpObjectFactory.nextObject();

		// find the matching public key encrypted data packet.
		PGPPublicKeyEncryptedData pgpPublicKeyEncryptedData = null;

		// build decryptor factory
		PublicKeyDataDecryptorFactory dataDecryptorFactory = null;
		for (Object pgpEncData : pgpEncryptedDataList) {
			PGPPublicKeyEncryptedData pkEnc = (PGPPublicKeyEncryptedData) pgpEncData;
			// for (PGPPrivateKey privateKey : privateKeys) {
			if (pkEnc.getKeyID() == privateKey.getKeyID()) {
				pgpPublicKeyEncryptedData = pkEnc;
				dataDecryptorFactory = new JcePublicKeyDataDecryptorFactoryBuilder().setProvider("BC")
						.build(privateKey);
				break;
			}
			// }
		}

		if (pgpPublicKeyEncryptedData == null || dataDecryptorFactory == null) {
			throw new IllegalStateException("matching encrypted data not found");
		}

		InputStream clear = pgpPublicKeyEncryptedData.getDataStream(dataDecryptorFactory);
//#4 Get the plain text by using your private key to decrypt the crypto text
		PGPObjectFactory plainFact = new PGPObjectFactory(clear, new JcaKeyFingerprintCalculator());
		Object message = plainFact.nextObject();

		ByteArrayOutputStream actualOutput = new ByteArrayOutputStream();
		Streams.pipeAll(((PGPLiteralData) message).getInputStream(), actualOutput);
		actualOutput.close();
		clear.close();
		// check data decrypts okay
		if (!pgpPublicKeyEncryptedData.verify()) {
			throw new RuntimeException("Enc data is invalid!");
		}
		return new String(actualOutput.toByteArray());
	}
}
